@@ -15,7 +15,7 @@
 # See the COPYING file for license information.
 #
 # Copyright (c) 2011,2012 peo3 <peo314159265@gmail.com>
-
+import sys
 import os
 import os.path
 import re
@@ -24,7 +24,11 @@ import errno
 
 from cgutils import host
 from cgutils import process
-import fileops
+from . import fileops
+
+
+if sys.version_info.major == 3:
+    long = int
 
 
 class SubsystemStatus(dict):
@@ -115,14 +119,14 @@ class SubsystemStatus(dict):
         self._update()
 
     def get_all(self):
-        return self.keys()
+        return list(self.keys())
 
     def get_available(self):
-        return [name for name in self.keys()
+        return [name for name in list(self.keys())
                 if self[name]['enabled']]
 
     def get_enabled(self):
-        return self.paths.keys()
+        return list(self.paths.keys())
 
     def get_path(self, subsys):
         return self.paths[subsys]
@@ -317,11 +321,14 @@ class SubsystemCpuacct(Subsystem):
 class SubsystemCpuset(Subsystem):
     NAME = 'cpuset'
     STATS = {
+        # str object something like '0', '0-1', and '0-1,3,4'
+        'effective_cpus': str,
+        'effective_mems': str,
         'memory_pressure': long,
     }
     CONFIGS = {
         'cpu_exclusive': 0,
-        # str object something like '0', '0-1', and '0-1,3,4'
+        # same as 'effective_*' ones
         'cpus': host.CPUInfo().get_online(),
         'mem_exclusive': 0,
         'mem_hardwall': 0,
@@ -375,6 +382,7 @@ class SubsystemMemory(Subsystem):
     }
     CONTROLS = {
         'force_empty': None,
+        'pressure_level': None,
     }
 
 
@@ -382,15 +390,23 @@ class SubsystemBlkio(Subsystem):
     NAME = 'blkio'
     STATS = {
         'io_merged': BlkioStat,
+        'io_merged_recursive': BlkioStat,
         'io_queued': BlkioStat,
+        'io_queued_recursive': BlkioStat,
         'io_service_bytes': BlkioStat,
+        'io_service_bytes_recursive': BlkioStat,
         'io_service_time': BlkioStat,
+        'io_service_time_recursive': BlkioStat,
         'io_serviced': BlkioStat,
+        'io_serviced_recursive': BlkioStat,
         'io_wait_time': BlkioStat,
+        'io_wait_time_recursive': BlkioStat,
         'sectors': SimpleStat,
+        'sectors_recursive': SimpleStat,
         'throttle.io_service_bytes': BlkioStat,
         'throttle.io_serviced': BlkioStat,
         'time': SimpleStat,
+        'time_recursive': SimpleStat,
         # Debugging files (Appeared only CONFIG_DEBUG_BLK_CGROUP=y)
         'avg_queue_size': BlkioStat,
         'dequeue': BlkioStat,
@@ -400,6 +416,8 @@ class SubsystemBlkio(Subsystem):
         'unaccounted_time': BlkioStat,
     }
     CONFIGS = {
+        'leaf_weight': 1000,
+        'leaf_weight_device': SimpleStat({}),
         'throttle.read_iops_device': SimpleStat({}),
         'throttle.write_iops_device': SimpleStat({}),
         'throttle.read_bps_device': SimpleStat({}),
@@ -446,7 +464,7 @@ class SubsystemNetPrio(Subsystem):
     }
     __ifs = os.listdir('/sys/class/net')
     CONFIGS = {
-        'ifpriomap': SimpleStat(zip(__ifs, [0] * len(__ifs))),
+        'ifpriomap': SimpleStat(list(zip(__ifs, [0] * len(__ifs)))),
     }
 
 
@@ -494,11 +512,11 @@ def _get_subsystem(name):
     return _subsystem_name2class[name]()
 
 
-class NoSuchControlFileError(StandardError):
+class NoSuchControlFileError(Exception):
     pass
 
 
-class IsRootGroupError(StandardError):
+class IsRootGroupError(Exception):
     pass
 
 
@@ -577,9 +595,9 @@ class CGroup:
             self.parent = get_cgroup(os.path.dirname(self.fullpath))
 
         self.paths = {}
-        for file in self._STATS.keys() + self._CONFIGS.keys() + self._CONTROLS.keys():
+        for file in list(self._STATS.keys()) + list(self._CONFIGS.keys()) + list(self._CONTROLS.keys()):
             self.paths[file] = os.path.join(self.fullpath, file)
-        for file in subsystem.STATS.keys() + subsystem.CONFIGS.keys() + subsystem.CONTROLS.keys():
+        for file in list(subsystem.STATS.keys()) + list(subsystem.CONFIGS.keys()) + list(subsystem.CONTROLS.keys()):
             self.paths[file] = os.path.join(self.fullpath, subsystem.name + '.' + file)
 
         self.configs = {}
@@ -629,13 +647,13 @@ class CGroup:
         which are categorised in the configs group.
         """
         configs = {}
-        for name, default in self.configs.iteritems():
+        for name, default in self.configs.items():
             cls = default.__class__
             path = self.paths[name]
             if os.path.exists(path):
                 try:
                     configs[name] = self._PARSERS[cls](fileops.read(path))
-                except IOError, e:
+                except IOError as e:
                     if e.errno == errno.EOPNOTSUPP:
                         # Since 3.5 memory.memsw.* are always created even if disabled.
                         # If disabled we will get EOPNOTSUPP when read or write them.
@@ -658,12 +676,12 @@ class CGroup:
         which are categorised in the stats group.
         """
         stats = {}
-        for name, cls in self.stats.iteritems():
+        for name, cls in self.stats.items():
             path = self.paths[name]
             if os.path.exists(path):
                 try:
                     stats[name] = self._PARSERS[cls](fileops.read(path))
-                except IOError, e:
+                except IOError as e:
                     # XXX: we have to distinguish unexpected errors from the expected ones
                     if e.errno == errno.EOPNOTSUPP:
                         # Since 3.5 memory.memsw.* are always created even if disabled.
@@ -693,7 +711,7 @@ class CGroup:
         new = get_cgroup(new_path)
         if set_initparams:
             params = self.subsystem.get_init_parameters(self.get_configs())
-            for filename, value in params.iteritems():
+            for filename, value in params.items():
                 new.set_config(filename, value)
         return new
 
@@ -731,6 +749,7 @@ class EventListener:
         'memory.usage_in_bytes',
         'memory.oom_control',
         'memory.memsw.usage_in_bytes',
+        'memory.pressure_level',
     ]
 
     def __init__(self, cgroup, target_name):
@@ -763,6 +782,9 @@ class EventListener:
         if target_name in ['memory.usage_in_bytes', 'memory.memsw.usage_in_bytes']:
             threshold = arguments[0]
             line = "%d %d %d\0" % (self.event_fd, self.target_fd, long(threshold))
+        elif target_name in ['memory.pressure_level']:
+            threshold = arguments[0]
+            line = "%d %d %s\0" % (self.event_fd, self.target_fd, threshold)
         else:
             line = "%d %d\0" % (self.event_fd, self.target_fd)
         os.write(self.ec_fd, line)
@@ -777,7 +799,7 @@ class EventListener:
 
 
 def _scan_cgroups_recursive(subsystem, fullpath, mount_point, filters):
-    cgroup = CGroup(subsystem, fullpath, filters)
+    cgroup = CGroup(subsystem, fullpath=fullpath, filters=filters)
 
     _childs = []
     for _file in os.listdir(fullpath):
@@ -793,7 +815,7 @@ def _scan_cgroups_recursive(subsystem, fullpath, mount_point, filters):
 #
 #  Public APIs
 #
-class NoSuchSubsystemError(StandardError):
+class NoSuchSubsystemError(Exception):
     pass
 
 
@@ -837,11 +859,11 @@ def get_cgroup(fullpath):
 
     status = SubsystemStatus()
     name = None
-    for name, path in status.paths.iteritems():
+    for name, path in status.paths.items():
         if path in fullpath:
             break
     else:
-        raise StandardError('Invalid path: ' + fullpath)
+        raise Exception('Invalid path: ' + fullpath)
     subsys = _get_subsystem(name)
 
     return CGroup(subsys, fullpath)
